#!/usr/bin/env python3
import re
import os
import hmac
import json
import time
import uuid
import base64
import random
import string
import hashlib
import datetime
import requests
import urllib.parse

from typing import Union

from Cryptodome.PublicKey import RSA
from Cryptodome.Cipher import AES, PKCS1_v1_5
from Cryptodome.Random import get_random_bytes

from .config import DEVICE, DEVICE_LIST, COUNTRY, COUNTRY_LIST, COUNTRY_NAME, THREADS_RELEASE, INSTAGRAM_RELEASE

class Generator:
    
    def uuid(self, heex: bool = False, seed: Union[str, dict] = None, upper: bool = False):
        if seed is not None:
            hash = hashlib.md5()
            hash.update(seed.encode('utf-8'))
            _uid = uuid.UUID(hash.hexdigest())
        else:
            _uid = uuid.uuid4()
        if heex: return str(_uid.hex).upper() if upper else str(_uid.hex)
        return str(_uid).upper() if upper else str(_uid)
    
    def nonce(self, seed: str = None):
        seed = seed if seed is not None else '12345'
        seed = base64.b64encode(hashlib.sha256(seed.encode('utf-8')).digest())
        return seed.decode('utf-8')
    
    def string(self, length: int = 16, with_char: bool = False):
        base_char = string.ascii_uppercase + string.ascii_lowercase + string.digits
        return ''.join(random.choice(base_char + '-_' if with_char else base_char) for _ in range(length))
    
    def jazoest(self, seed: str = None):
        seed = seed if seed is not None else self.android_id()
        return f'2{sum(ord(line) for line in seed)}'
    
    def android_id(self, device_id: any = None):
        device_id = device_id if device_id is not None else self.uuid()
        seed = str(device_id).replace('-','')
        hash = hashlib.sha256(seed.encode())
        return 'android-' + hash.hexdigest()[:16]
    
    def machine_id(self, length: int = 28, with_char: bool = False):
        base_char = string.ascii_uppercase + string.ascii_lowercase + string.digits
        return ''.join(random.choice(base_char + '-_' if with_char else base_char) for _ in range(length))
    
    def device_id(self):
        return self.uuid(False)
    
    def family_device_id(self, upper: bool = False):
        return self.uuid(upper=upper)
    
    def pigeon_session_id(self):
        return 'UFS-' + self.uuid(False) + '-' + random.choice(['0', str(random.randint(1,6))])
    
    def pigeon_raw_client_time(self):
        return str(round(time.time(), 3))
    
    def timestamp(self):
        return datetime.datetime.now().timestamp()
    
    def timestamp_to_datetime(self, timestamp: Union[str, int]):
        date = datetime.datetime.fromtimestamp(timestamp)
        return date.strftime('%Y-%m-%d %H:%M:%S')
    
    def wordlist(self, username: str = None, fullname: str = None, combolist: list = ['123','12345'], shuffle: bool = False, capitalize: bool = False):
        wordlist = []
        digit = ''.join([str(line) for line in username if line.isdigit()])
        users = username.replace(digit,'').replace('.','').replace('_','')
        names = ''.join(line.lower() for line in fullname if line.isalpha() or line.isspace())
        if len(digit) > 0: combolist.append(digit)
        for combo in combolist:
            for name in names.split(' '):
                if len(name) > 3: wordlist.append(name + combo)
            wordlist.append(users.replace(' ','') + combo)
            wordlist.append(names.replace(' ','') + combo)
        wordlist = [word for word in wordlist if all(ord(line) < 128 for line in word)]
        wordlist = [word.replace(' ','') for word in wordlist if len(word) >= 8]
        wordlist = sorted(wordlist, key=len)
        if shuffle: random.shuffle(wordlist)
        if capitalize: wordlist = [line.capitalize() for line in wordlist]
        return wordlist
    
    def signature(self, data: dict = None, ig_sig_key: str = 'SIGNATURE', ig_sig_key_version: str = None):
        if ig_sig_key.isdigit():
            return {
                'signed_body': hmac.new(ig_sig_key.encode('utf-8'), json.dumps(data).encode('utf-8'), hashlib.sha256).hexdigest() + '.' + json.dumps(data),
                'ig_sig_key_version': ig_sig_key_version
            }
        else:
            if ig_sig_key_version is not None:
                return {
                    'signed_body': ig_sig_key + '.' + json.dumps(data),
                    'ig_sig_key_version': ig_sig_key_version
                }
            return {
                'signed_body': ig_sig_key + '.' + json.dumps(data)
            }
    
    def cookie_string(self, cookie: dict = None):
        if not 'mid' in cookie: cookie['mid'] = self.machine_id()
        if not 'ig_did' in cookie: cookie['ig_did'] = self.device_id(True)
        return '; '.join([key + '=' + value for key, value in cookie.items()]) if cookie is not None else ''
    
    def encrypt_bearer(self, cookie: str = None):
        try: return 'Bearer IGT:2:{}'.format(base64.b64encode(str({'ds_user_id': re.search('ds_user_id=(.*?);', str(cookie)).group(1), 'sessionid': re.search('sessionid=(.*?);', str(cookie) + ';').group(1), 'should_use_header_over_cookies': True}).replace("'",'"').replace('True','true').replace(' ','').encode('ascii')).decode('ascii'))
        except: return None
    
    def decrypt_bearer(self, bearer: str = None):
        cookie = json.loads(base64.urlsafe_b64decode(bearer.split(':')[-1]).decode('utf-8'))
        try: cookie.pop('should_use_header_over_cookies')
        except: pass
        return self.cookie_string(cookie)
    
    def encrypt_password(self, password: str = None, timestamp: str = None, public_key: str = None, public_key_id: int = None):
        iv = get_random_bytes(12)
        session_key = get_random_bytes(32)
        public_key_id = public_key_id or 41
        timestamp = timestamp if timestamp is not None else str(self.timestamp())[:10]
        recipient_key = RSA.import_key(public_key if public_key is not None else '-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAvcu1KMDR1vzuBr9iYKW8\nKWmhT8CVUBRkchiO8861H7zIOYRwkQrkeHA+0mkBo3Ly1PiLXDkbKQZyeqZbspke\n4e7WgFNwT23jHfRMV/cNPxjPEy4kxNEbzLET6GlWepGdXFhzHfnS1PinGQzj0ZOU\nZM3pQjgGRL9fAf8brt1ewhQ5XtpvKFdPyQq5BkeFEDKoInDsC/yKDWRAx2twgPFr\nCYUzAB8/yXuL30ErTHT79bt3yTnv1fRtE19tROIlBuqruwSBk9gGq/LuvSECgsl5\nz4VcpHXhgZt6MhrAj6y9vAAxO2RVrt0Mq4OY4HgyYz9Wlr1vAxXXGAAYIvrhAYLP\n7QIDAQAB\n-----END PUBLIC KEY-----\n')
        cipher_rsa = PKCS1_v1_5.new(recipient_key)
        rsa_encrypted = cipher_rsa.encrypt(session_key)
        cipher_aes = AES.new(session_key, AES.MODE_GCM, iv)
        cipher_aes.update(timestamp.encode())
        aes_encrypted, tag = cipher_aes.encrypt_and_digest(password.encode('utf-8'))
        size_buffer = len(rsa_encrypted).to_bytes(2, byteorder='little')
        payload = base64.b64encode(b''.join([b'\x01',public_key_id.to_bytes(1, byteorder='big'),iv,size_buffer,rsa_encrypted,tag,aes_encrypted]))
        return f'#PWD_INSTAGRAM:4:{timestamp}:{payload.decode()}'

class Device:
    
    def __init__(self, device_brand: str = None, device_model: str = None, device_country: str = None, device_operator: str = None):
        self.device_brand = device_brand
        self.device_model = device_model
        self.device_country = device_country
        self.device_operator = device_operator
        self.device_data = None
        self.country_data = None
        
    def device_dpi(self):
        return random.choice(['480dpi; 1080x2400','480dpi; 720x1600','480dpi; 720x1560','480dpi; 1080x2376','480dpi; 1080x2404','480dpi; 1080x2408','320dpi; 1080x2340','560dpi; 1440x3040','560dpi; 1440x3088','560dpi; 1080x2400','320dpi; 1600x2560','320dpi; 720x1568','560dpi; 1440x2560','480dpi; 1344x2772'])
    
    def device_build(self):
        return '{}.{}'.format(random.choice(['SP1A','QP1A','RP1A','TP1A','RKQ1','SKQ1']), str(random.randint(200999,220905)) + '.0{}'.format(random.choice(['01','02','03','04','05','06','07','08','09','10','11','12','13','14','15','16'])))
    
    def device_armeabi(self):
        return random.choice(['arm64-v8a','armeabi-v8a:armeabi','armeabi-v7a:armeabi','x86_64:x86:x86_64','x86_64:arm64-v8a:x86_64'])
    
    def device_density(self):
        return random.choice(['{density=3.0,width=1080,height=2068}','{density=3.0,width=1080,height=1920}','{density=2.3,width=2149,height=1117}','{density=1.0,width=2060,height=1078}','{density=1.8,width=1582,height=558}','{density=3.0,width=1080,height=1920}','{density=2.0,width=720,height=1193}','{density=2.1,width=1814,height=1023}'])
    
    def info(self):
        if self.device_brand and self.device_brand.upper() in DEVICE_LIST:
            if self.device_model:
                self.device_data = next((d for d in DEVICE[self.device_brand.upper()] if d['device_model'] == self.device_model), None)
                if not self.device_data: self.device_data = random.choice(DEVICE[self.device_brand.upper()])
            else: self.device_data = random.choice(DEVICE[self.device_brand.upper()])
        else:
            self.device_brand = random.choice(DEVICE_LIST)
            self.device_data = random.choice(DEVICE[self.device_brand.upper()])
        
        if not self.device_country and self.device_country not in COUNTRY_LIST: self.device_country = random.choice(COUNTRY_LIST)
        self.country_data = COUNTRY[self.device_country.upper()]
        if not self.device_operator: self.device_operator = random.choice(COUNTRY[self.device_country]['operator'])
        return {
            'device_brand': self.device_brand.capitalize() if len(self.device_brand) > 3 else self.device_brand,
            'device_model': self.device_data['device_model'],
            'device_board': self.device_data['device_board'],
            'device_build': self.device_build(),
            'device_vendor': self.device_data['device_vendor'],
            'device_version': self.device_data['device_version'],
            'device_armeabi': self.device_armeabi(),
            'device_density': self.device_density(),
            'device_dpi': self.device_dpi(),
            'device_sdk': str(19 + int(self.device_data['device_version'])),
            'device_number': self.country_data['number'],
            'device_country': self.device_country,
            'device_language': self.country_data['language'],
            'device_operator': random.choice(self.country_data['operator'])
        }

class UserAgent:
    
    def __init__(self, device: dict = None):
        self.device = device if device is not None else Device().info()

    def chrome_version(self):
        return str(random.randint(80,112)) + '.0.' + str(random.randint(2024,3058)) + '.0' + str(random.randint(10,99))
    
    def threads_version(self):
        return random.choice(THREADS_RELEASE)
    
    def instagram_version(self):
        return random.choice(INSTAGRAM_RELEASE)
    
    def chrome(self, chrome_version: str = None, **kwargs):
        device = self.device.copy()
        device['chrome_version'] = chrome_version if chrome_version is not None else self.chrome_version()
        return 'Mozilla/5.0 (Linux; Android {device_version}; {device_model} Build/{device_build}) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{chrome_version} Mobile Safari/537.36'.format(**device)
    
    def threads(self, threads_code: str = None, threads_version: str = None, **kwargs):
        device = self.device.copy()
        threads = self.threads_version()
        device['threads_code'] = threads_code if threads_code is not None else threads['code']
        device['threads_version'] = threads_version if threads_version is not None else threads['version']
        return 'Barcelona {threads_version} Android ({device_sdk}/{device_version}; {device_dpi}; {device_brand}; {device_model}; {device_board}; {device_vendor}; {device_language}; {threads_code})'.format(**device)
    
    def instagram(self, instagram_code: str = None, instagram_version: str = None, **kwargs):
        device = self.device.copy()
        instagram = self.threads_version()
        device['instagram_code'] = instagram_code if instagram_code is not None else instagram['code']
        device['instagram_version'] = instagram_version if instagram_version is not None else instagram['version']
        return 'Instagram {instagram_version} Android ({device_sdk}/{device_version}; {device_dpi}; {device_brand}; {device_model}; {device_board}; {device_vendor}; {device_language}; {instagram_code})'.format(**device)
    
class Instagram:
    
    web = 'https://www.instagram.com'
    api = 'https://i.instagram.com/api/v1'
    
    def __init__(
            self,
            cookie: str = None,
            bearer: str = None,
            device: dict = None,
            device_id: str = None,
            csrftoken: str = None,
            machine_id: str = None,
            user_agent: str = None,
            session: requests.Session = None,
            **kwargs
        ):
        
        generator = Generator()
        
        if not cookie:
            raise Exception(
                'cookie cannot be empty '
                'require Instagram cookie to access Instagram private API'
            )

        try: 
            user_id = str(re.search(r'ds_user_id=(.*?);', cookie).group(1))
        except (AttributeError, IndexError):
            raise Exception(
                'cookie is not valid '
                'cannot find ds_user_id in this cookie'
            )

        device = device or Device(device_country='ID').info()
        bearer = bearer or generator.encrypt_bearer(cookie)
        device_id = (
            str(uuid.UUID(re.search(r'ig_did=(.*?);', cookie).group(1)))
            if 'ig_did' in cookie else generator.device_id()
        )
        csrftoken = (
            re.search(r'csrftoken=(.*?);', cookie).group(1)
            if 'csrftoken' in cookie else generator.string(32)
        )
        machine_id = (
            re.search(r'mid=(.*?);', cookie).group(1)
            if 'mid' in cookie else generator.machine_id()
        )
        
        user_agent = user_agent or UserAgent(device).instagram()
        attributes = {
            'cookie': cookie,
            'bearer': bearer,
            'device': device,
            'user_id': user_id,
            'device_id': device_id,
            'csrftoken': csrftoken,
            'machine_id': machine_id,
            'android_id': generator.android_id(device_id),
            'user_agent': user_agent,
            'session': session or requests.Session(),
            'kwargs': kwargs
        }

        for key, value in attributes.items():
            setattr(self, key, value)
        
        self.session.cookies.update({'cookie': self.cookie})
        self.session.headers.update({
            'Host': 'i.instagram.com',
            'Authorization': self.bearer,
            'Accept-Encoding': 'gzip, deflate',
            'Accept-Language': '{}, en-US'.format(self.device.get('device_language').replace('_','-')),
            'X-Ig-App-Id': '567067343352427',
            'X-Ig-Device-ID': self.device_id,
            'X-Ig-Android-ID': self.android_id,
            'X-Ig-Connection-Type': 'MOBILE(LTE)',
            'X-Fb-Connection-Type': 'MOBILE.LTE',
            'X-Fb-Http-Engine': 'Liger',
            'X-MID': self.machine_id,
            'User-Agent': self.user_agent
        })
    
    def download(self, url: str = None, name: str = None, path: str = None):
        if not path or not os.path.isdir(path): path = '.'
        filepath = os.path.join(path, name)
        with requests.get(url, stream=True) as session:
            session.raise_for_status()
            with open(filepath, 'wb') as files:
                for chunk in session.iter_content(chunk_size=8192): 
                    files.write(chunk)
        if os.path.isfile(filepath): return filepath
        else: return False
    
    def account(self):
        try:
            user = self.session.get(self.api + f'/users/{self.user_id}/info/').json()['user']
            info = {'id': user['pk_id'], 'email': '', 'phone': '', 'private': user['is_private'], 'verified': user['is_verified'], 'username': user['username'], 'fullname': user['full_name'], 'birthday': '', 'followers': str(user['follower_count']), 'following': str(user['following_count']), 'mediapost': str(user['media_count']), 'biography': user['biography'], 'pictures': user['hd_profile_pic_url_info']['url']}
            user = self.session.get(self.api + '/accounts/current_user/?edit=true').json()['user']
            info['email'] = user['email'] or ''
            info['phone'] = user['phone_number'] or ''
            info['birthday'] = user['birthday'] or ''
            return info
        except:
            return False
    
    def username_info(self, username: str = None):
        try:
            user = self.session.get(self.api + f'/users/{username}/usernameinfo/').json()['user']
            info = {'id': user['pk_id'], 'private': user['is_private'], 'verified': user['is_verified'], 'username': user['username'], 'fullname': user['full_name'], 'followers': str(user['follower_count']), 'following': str(user['following_count']), 'mediapost': str(user['media_count']), 'biography': user['biography'], 'pictures': user['hd_profile_pic_url_info']['url']}
            return info
        except:
            return False
    
    def followers(self, username: str = None):
        if not username.isdigit(): id = self.username_info(username=username)['id']
        else: id = username
        end_cursor = None
        has_next_page = True
        while has_next_page:
            try:
                response = self.session.get(self.web + '/graphql/query/', params={'query_hash': 'c76146de99bb02f6415203be841dd25a', 'id': id, 'first': 100, 'after': end_cursor}).json()
                has_next_page = response['data']['user']['edge_followed_by']['page_info']['has_next_page']
                for user in response['data']['user']['edge_followed_by']['edges']:
                    try:
                        data = {'id': user['node']['id'], 'username': user['node']['username'], 'fullname': user['node']['full_name']}
                        yield data
                    except (requests.exceptions.JSONDecodeError, json.decoder.JSONDecodeError): continue
                if has_next_page: end_cursor = response['data']['user']['edge_followed_by']['page_info']['end_cursor']
            except: break
    
    def following(self, username: str = None):
        if not username.isdigit(): id = self.username_info(username=username)['id']
        else: id = username
        end_cursor = None
        has_next_page = True
        while has_next_page:
            try:
                response = self.session.get(self.web + '/graphql/query/', params={'query_hash': 'd04b0a864b4b54837c0d870b0e77e076', 'id': id, 'first': 100, 'after': end_cursor}).json()
                has_next_page = response['data']['user']['edge_follow']['page_info']['has_next_page']
                for user in response['data']['user']['edge_follow']['edges']:
                    try:
                        data = {'id': user['node']['id'], 'username': user['node']['username'], 'fullname': user['node']['full_name']}
                        yield data
                    except (requests.exceptions.JSONDecodeError, json.decoder.JSONDecodeError): continue
                if has_next_page: end_cursor = response['data']['user']['edge_follow']['page_info']['end_cursor']
            except: break
    
    def media_id(self, url: str = None):
        try:
            media = 0
            base_char = string.ascii_uppercase + string.ascii_lowercase + string.digits + '-_'
            if '/p/' in url: code = re.search(r"/p/([A-Za-z0-9_-]+)/\?", url).group(1)
            elif '/reel/' in url: code = re.search(r"/reel/([A-Za-z0-9_-]+)/\?", url).group(1)
            elif '/story/' in url: code = re.search(r"/story/([A-Za-z0-9_-]+)/\?", url).group(1)
            else: code = ''
            for line in code: media = (media * 64) + base_char.index(line)
            return media
        except:
            return False
    
    def media_info(self, url: str = None):
        try:
            if not url.isdigit(): media_id = self.media_id(url)
            else: media_id = url
            media = []
            media_type = lambda x: 'image' if x == 1 else ('video' if x == 2 else 'carousel')
            items = self.session.get(self.api + f'/media/{media_id}/info/').json()['items']
            for line in items:
                try: caption = line['caption']['text']
                except: caption = ''
                try: id = line['caption']['user']['id']
                except: id = ''
                try: username = line['caption']['user']['username']
                except: username = ''
                try: fullname = line['caption']['user']['full_name']
                except: fullname = ''
                try: has_more_comments = line['has_more_comments']
                except: has_more_comments = False
                try:
                    if media_type(line['media_type']) == 'image':
                        media_url = line['image_versions2']['candidates'][0]['url']
                        media_name = re.search(r'([^/]+\.(jpg|png))', media_url).group(1)
                        media.append({'url': media_url, 'name': media_name})
                    elif media_type(line['media_type']) == 'video':
                        media_url = line['video_versions'][0]['url']
                        media_name = re.search(r'([^/]+\.(mp4))', media_url).group(1)
                        media.append({'url': media_url, 'name': media_name})
                    elif media_type(line['media_type']) == 'carousel':
                        for lines in line['carousel_media']:
                            if media_type(lines['media_type']) == 'image':
                                media_url = lines['image_versions2']['candidates'][0]['url']
                                media_name = re.search(r'([^/]+\.(jpg|png))', media_url).group(1)
                                media.append({'url': media_url, 'name': media_name})
                            elif media_type(lines['media_type']) == 'video': 
                                media_url = lines['video_versions'][0]['url']
                                media_name = re.search(r'([^/]+\.(mp4))', media_url).group(1)
                                media.append({'url': media_url, 'name': media_name})
                except: pass
                return {
                    'id': id,
                    'username': username,
                    'fullname': fullname,
                    'taken_at': generator.timestamp_to_datetime(line['taken_at']),
                    'like': str(line['like_count']),
                    'liked': line['has_liked'],
                    'comment': str(line['comment_count']),
                    'caption': caption,
                    'can_save': line['can_viewer_save'],
                    'can_share': line['can_viewer_reshare'],
                    'can_comment': has_more_comments,
                    'media_id': line['id'],
                    'media_type': media_type(line['media_type']),
                    'media_code': line['code'],
                    'media': media}
        except:
            return False
    
    def upload_id(self, file: str = None, cookie: dict = None, headers: dict = None):
        if not os.path.isfile(file): raise FileNotFoundError(file)
        upload_id = str(int(time.time()) * 1000)
        entity_name = '{}_0_{}'.format(upload_id, random.randint(1000000000, 9999999999))
        params = {'retry_context':'{"num_step_auto_retry": 0, "num_reupload": 0, "num_step_manual_retry": 0}"','media_type':'1','xsharing_user_ids':'[]','upload_id': upload_id,'image_compression':json.dumps({'lib_name':'moz','lib_version':'3.1.m','quality': 80})}
        with open(file,'rb') as f:
            file_data = f.read()
            file_size = str(len(file_data))
        session = requests.Session()
        session.cookies.update(cookie)
        session.headers.update(headers)
        session.headers.update({'Accept-Encoding':'gzip','X-Instagram-Rupload-Params':json.dumps(params),'X_FB_PHOTO_WATERFALL_ID':generator.uuid(),'X-Entity-Type':'image/jpeg','Offset':'0','X-Entity-Name':entity_name,'X-Entity-Length':file_size,'Content-Type':'application/octet-stream','Content-Length':file_size})
        upload_id = session.post('https://i.instagram.com/rupload_igphoto/{}'.format(entity_name), data=file_data).json()['upload_id']
        return upload_id
    
    def change_biography(self, text: str = None):
        try:
            response = self.session.post(
                self.api + '/accounts/set_biography/',
                data = {
                    'logged_in_uids': self.user_id,
                    'device_id': self.device_id,
                    '__uuid': generator.uuid(),
                    'raw_text': text,
                }
            ).json()
            if response['status'] == 'ok': return True
            else: return False
        except:
            return False
    
    def change_profile_picture(self, file: str = None):
        try:
            if not os.path.isfile(file): raise FileNotFoundError(file)
            upload_id = self.upload_id(file=file, cookie=self.session.cookies.get_dict().copy(), headers=self.session.headers.copy())
            response = self.session.post(
                self.api + '/accounts/change_profile_picture/',
                data = {
                    '_uuid': generator.uuid(),
                    'use_fbuploader': False,
                    'remove_birthday_selfie': False,
                    'upload_id': upload_id
                }
            ).json()
            if response['status'] == 'ok': return True
            else: return False
        except:
            return False
    
    def remove_profile_picture(self):
        try:
            response = self.session.post(
                self.api + '/accounts/remove_profile_picture/',
                data = {
                    '_uuid': generator.uuid(),
                    'use_fbuploader': False,
                    'remove_birthday_selfie': False,
                    'upload_id': None
                }
            ).json()
            if response['status'] == 'ok': return True
            else: return False
        except:
            return False
    
    def change_to_public(self):
        try:
            response = self.session.post(self.api + '/accounts/set_public/').json()
            if response['status'] == 'ok': return True
            else: return False
        except:
            return False
    
    def change_to_private(self):
        try:
            response = self.session.post(self.api + '/accounts/set_private/').json()
            if response['status'] == 'ok': return True
            else: return False
        except:
            return False
    
    def change_password(self, old_password: str = None, new_password: str = None):
        try:
            enc_old_password = self.encrypt_password(old_password)
            enc_new_password = self.encrypt_password(new_password)
            response = self.session.post(
                self.api + '/accounts/change_password/',
                data = {
                    '_uid': self.user_id,
                    '_uuid': self.device_id,
                    '_csrftoken': self.csrftoken,
                    'enc_old_password': enc_old_password,
                    'enc_new_password1': enc_new_password,
                    'enc_new_password2': enc_new_password
                }
            ).json()
            if response['status'] == 'ok': return True
            else: return False
        except:
            return False
    
    def like(self, url: str = None):
        try:
            if not url.isdigit(): media_id = self.media_id(url)
            else: media_id = url
            response = self.session.post(
                self.api + f'/media/{media_id}/like/',
                data = {
                    'media_id': media_id
                }
            ).json()
            if response['status'] == 'ok': return True
            else: return False
        except:
            return False
    
    def unlike(self, url: str = None):
        try:
            if not url.isdigit(): media_id = self.media_id(url)
            else: media_id = url
            response = self.session.post(
                self.api + f'/media/{media_id}/unlike/',
                data = {
                    'media_id': media_id
                }
            ).json()
            if response['status'] == 'ok': return True
            else: return False
        except:
            return False
    
    def block(self, username: str = None):
        try:
            if not username.isdigit(): id = self.username_info(username=username)['id']
            else: id = username
            response = self.session.post(
                self.api + f'/friendships/block/{id}/',
                data = {
                    'surface': 'profile',
                    'is_auto_block_enabled': 'true',
                    'user_id':id,
                    '_uid': self.user_id,
                    '__uuid': self.device_id,
                    'container_module': 'profile'
                }
            ).json()
            if response['status'] == 'ok': return True
            else: return False
        except:
            return False
    
    def unblock(self, username: str = None):
        try:
            if not username.isdigit(): id = self.username_info(username=username)['id']
            else: id = username
            response = self.session.post(
                self.api + f'/friendships/unblock/{id}/',
                data = {
                    'user_id': id,
                    '_uid': self.user_id,
                    '__uuid': self.device_id,
                    'container_module': 'search_typeahead'
                }
            ).json()
            if response['status'] == 'ok': return True
            else: return False
        except:
            return False
    
    def follow(self, username: str = None):
        try:
            if not username.isdigit(): id = self.username_info(username=username)['id']
            else: id = username
            response = self.session.post(
                self.api + f'/friendships/create/{id}/',
                data = {
                    'user_id': id
                }
            ).json()
            if response['status'] == 'ok': return True
            else: return False
        except:
            return False
    
    def unfollow(self, username: str = None):
        try:
            if not username.isdigit(): id = self.username_info(username=username)['id']
            else: id = username
            response = self.session.post(
                self.api + f'/friendships/destroy/{id}/',
                data = {
                    'user_id': id
                }
            ).json()
            if response['status'] == 'ok': return True
            else: return False
        except:
            return False
    
    def comment(self, url: str = None, text: str = None):
        try:
            if not url.isdigit(): media_id = self.media_id(url)
            else: media_id = url
            text = text if text is not None else random.choice(['Nice post! 👍', 'Good work bro! 💯', 'Alatnya bagus banget, gw suka! 😍', 'Gw kira boongan, ternyata work cuy! 🔥', 'Mantap banget ini alat! 👏', 'Keren banget, auto recommend! 🚀', 'Wih, ini beneran ngebantu banget! 🙌', 'Nice banget, langsung work! 😎', 'Good job! Alatnya oke punya! 💪', 'Ini alat beneran nggak ngecewain! 😁', 'Wah, keren banget! Auto save! 📂', 'Nice, langsung coba dan puas! 😄', 'Alatnya simpel tapi powerful! 💥', 'Gw suka banget sama ini! 👍', 'Ini beneran worth it, cuy! 💸', 'Good work, alatnya beneran ngebantu! 🙏', 'Wih, langsung work tanpa ribet! 🎉', 'Nice post, beneran informatif! 📚', 'Alatnya beneran bagus, gw puas! 😊', 'Keren banget, auto jadi favorit! ⭐', 'Ini beneran nggak boong, work banget! 🔥', 'Good job, alatnya beneran ngebantu! 💯', 'Wah, langsung suka sama ini! 😍', 'Nice, beneran sesuai ekspektasi! 😎', 'Alatnya beneran nggak ngecewain! 👌', 'Ini beneran keren, auto recommend! 🚀', 'Good work, langsung work tanpa error! 🙌', 'Wih, beneran ngebantu banget! 💪', 'Nice post, langsung coba dan puas! 😄', 'Alatnya beneran simpel tapi powerful! 💥', 'Gw suka banget sama ini, beneran worth it! 💸', 'Ini beneran nggak boong, work banget! 🔥', 'Good job, alatnya beneran ngebantu! 🙏', 'Wah, langsung suka sama ini! 😍', 'Nice, beneran sesuai ekspektasi! 😎', 'Alatnya beneran nggak ngecewain! 👌', 'Ini beneran keren, auto recommend! 🚀', 'Good work, langsung work tanpa error! 🙌', 'Wih, beneran ngebantu banget! 💪', 'Nice post, langsung coba dan puas! 😄', 'Alatnya beneran simpel tapi powerful! 💥', 'Gw suka banget sama ini, beneran worth it! 💸', 'Ini beneran nggak boong, work banget! 🔥', 'Good job, alatnya beneran ngebantu! 🙏', 'Wah, langsung suka sama ini! 😍', 'Nice, beneran sesuai ekspektasi! 😎', 'Alatnya beneran nggak ngecewain! 👌', 'Ini beneran keren, auto recommend! 🚀', 'Good work, langsung work tanpa error! 🙌', 'Wih, beneran ngebantu banget! 💪', 'Nice post, langsung coba dan puas! 😄', 'Alatnya beneran simpel tapi powerful! 💥', 'Gw suka banget sama ini, beneran worth it! 💸', 'Ini beneran nggak boong, work banget! 🔥', 'Good job, alatnya beneran ngebantu! 🙏', 'Wah, langsung suka sama ini! 😍', 'Nice, beneran sesuai ekspektasi! 😎', 'Alatnya beneran nggak ngecewain! 👌', 'Ini beneran keren, auto recommend! 🚀', 'Good work, langsung work tanpa error! 🙌', 'Wih, beneran ngebantu banget! 💪', 'Nice post, langsung coba dan puas! 😄', 'Alatnya beneran simpel tapi powerful! 💥', 'Gw suka banget sama ini, beneran worth it! 💸', 'Ini beneran nggak boong, work banget! 🔥', 'Good job, alatnya beneran ngebantu! 🙏', 'Wah, langsung suka sama ini! 😍', 'Nice, beneran sesuai ekspektasi! 😎', 'Alatnya beneran nggak ngecewain! 👌', 'Ini beneran keren, auto recommend! 🚀', 'Good work, langsung work tanpa error! 🙌', 'Wih, beneran ngebantu banget! 💪', 'Nice post, langsung coba dan puas! 😄', 'Alatnya beneran simpel tapi powerful! 💥', 'Gw suka banget sama ini, beneran worth it! 💸', 'Ini beneran nggak boong, work banget! 🔥', 'Good job, alatnya beneran ngebantu! 🙏', 'Wah, langsung suka sama ini! 😍', 'Nice, beneran sesuai ekspektasi! 😎', 'Alatnya beneran nggak ngecewain! 👌', 'Ini beneran keren, auto recommend! 🚀', 'Good work, langsung work tanpa error! 🙌', 'Wih, beneran ngebantu banget! 💪', 'Nice post, langsung coba dan puas! 😄', 'Alatnya beneran simpel tapi powerful! 💥', 'Gw suka banget sama ini, beneran worth it! 💸', 'Ini beneran nggak boong, work banget! 🔥', 'Good job, alatnya beneran ngebantu! 🙏', 'Wah, langsung suka sama ini! 😍', 'Nice, beneran sesuai ekspektasi! 😎', 'Alatnya beneran nggak ngecewain! 👌', 'Ini beneran keren, auto recommend! 🚀', 'Good work, langsung work tanpa error! 🙌', 'Wih, beneran ngebantu banget! 💪', 'Nice post, langsung coba dan puas! 😄', 'Alatnya beneran simpel tapi powerful! 💥', 'Gw suka banget sama ini, beneran worth it! 💸', 'Ini beneran nggak boong, work banget! 🔥', 'Good job, alatnya beneran ngebantu! 🙏', 'Wah, langsung suka sama ini! 😍', 'Nice, beneran sesuai ekspektasi! 😎', 'Alatnya beneran nggak ngecewain! 👌', 'Ini beneran keren, auto recommend! 🚀', 'Good work, langsung work tanpa error! 🙌', 'Wih, beneran ngebantu banget! 💪'])
            response = self.session.post(
                self.api + f'/media/{media_id}/comment/',
                data = {
                    'comment_text': text
                }
            ).json()
            if response['status'] == 'ok': return True
            else: return False
        except:
            return False
    
    def direct_message(self, username: str = None, text: str = None):
        try:
            if not username.isdigit(): id = self.username_info(username=username)['id']
            else: id = username
            text = text if text is not None else 'Hai! pesan ini dikirim menggunakan instahack 👋🏻\n\nhttps://github.com/termuxhackers-id/instahack'
            data = {'action':'send_item','client_context':generator.uuid(),'recipient_users':'[['+id+']]','__uuid':self.device_id}
            if 'https' in str(text):
                path = 'link'
                data.update({'link_text': text})
            else:
                path = 'text'
                data.update({'text': text})
            response = self.session.post(
                self.api + f'/direct_v2/threads/broadcast/{path}/',
                data=data
            ).json()
            if response['status'] == 'ok': return True
            else: return False
        except:
            return False
    
    def direct_message_photo(self, username: str = None, file: str = None):
        try:
            if not os.path.isfile(file): raise FileNotFoundError(file)
            if not username.isdigit(): id = self.username_info(username=username)['id']
            else: id = username
            upload_id = self.upload_id(file=file, cookie=self.session.cookies.get_dict().copy(), headers=self.session.headers.copy())
            response = self.session.post(
                self.api + f'/direct_v2/threads/broadcast/configure_photo/',
                data = {
                    'action': 'send_item',
                    'send_attribution': 'inbox',
                    'client_context': generator.uuid(),
                    '__uuid': self.device_id,
                    'upload_id': upload_id,
                    'recipient_users':'[['+id+']]',
                    'allow_full_aspect_ratio':'true'
                }
            ).json()
            if response['status'] == 'ok': return True
            else: return False
        except:
            return False
    
    def upload_photo(self, file: str = None, caption: str = None):
        try:
            if not os.path.isfile(file): raise FileNotFoundError(file)
            upload_id = self.upload_id(file=file, cookie=self.session.cookies.get_dict().copy(), headers=self.session.headers.copy())
            response = self.session.post(
                self.api + '/media/configure/',
                data = {
                    '_uid': self.user_id,
                    '__uuid': self.device_id,
                    'device_id': self.device_id,
                    'custom_accessibility_caption': caption,
                    'caption': caption,
                    'upload_id': upload_id
                }
            ).json()
            if response['status'] == 'ok': return True
            else: return False
        except:
            return False
    
    def upload_photo_story(self, file: str = None, caption: str = None, resolution: tuple[int, int] = (1080, 1920)):
        try:
            if not os.path.isfile(file): raise FileNotFoundError(file)
            upload_id = self.upload_id(file=file, cookie=self.session.cookies.get_dict().copy(), headers=self.session.headers.copy())
            date = datetime.datetime.now()
            data = {
                "supported_capabilities_new": "[{\"name\":\"SUPPORTED_SDK_VERSIONS\",\"value\":\"131.0,132.0,133.0,134.0,135.0,136.0,137.0,138.0,139.0,140.0,141.0,142.0,143.0,144.0,145.0,146.0,147.0,148.0,149.0,150.0,151.0,152.0,153.0,154.0,155.0,156.0,157.0,158.0,159.0\"},{\"name\":\"FACE_TRACKER_VERSION\",\"value\":\"14\"},{\"name\":\"COMPRESSION\",\"value\":\"ETC2_COMPRESSION\"},{\"name\":\"gyroscope\",\"value\":\"gyroscope_enabled\"}]",
                "has_original_sound": "1",
                "camera_entry_point": "12",
                "original_media_type": "1",
                "camera_session_id": generator.uuid(),
                "date_time_digitalized": f"{date.year}:{date.month:02}:{date.day:02} {date.hour:02}:{date.minute:02}:{date.second:02}",
                "camera_model": self.device["device_model"],
                "scene_capture_type": "",
                "timezone_offset": (datetime.datetime.fromtimestamp(date.timestamp() * 1e-3) - datetime.datetime.utcfromtimestamp(date.timestamp() * 1e-3)).seconds,
                "client_shared_at": int(date.timestamp()),
                "story_sticker_ids": "",
                "configure_mode": "1",
                "source_type": "3",
                "camera_position": "front",
                "_uid": self.user_id,
                "device_id": self.device_id,
                "composition_id": generator.uuid(),
                "_uuid": generator.uuid(),
                "creation_surface": "camera",
                "can_play_spotify_audio": "1",
                "date_time_original": f"{date.year}:{date.month:02}:{date.day:02} {date.hour:02}:{date.minute:02}:{date.second:02}",
                "capture_type": "normal",
                "upload_id": upload_id,
                "client_timestamp": int(date.timestamp()),
                "private_mention_sharing_enabled": "1",
                "media_transformation_info": f"{{\"width\":\"{resolution[0]}\",\"height\":\"{resolution[1]}\",\"x_transform\":\"0\",\"y_transform\":\"0\",\"zoom\":\"1.0\",\"rotation\":\"0.0\",\"background_coverage\":\"0.0\"}}",
                "camera_make": self.device["device_brand"],
                "device": {
                    "manufacturer": self.device["device_brand"],
                    "model": self.device["device_model"],
                    "android_version": int(self.device["device_sdk"]),
                    "android_release": self.device["device_version"]
                },
                "edits": {
                    "filter_type": 0,
                    "filter_strength": 1.0,
                    "crop_original_size": [
                        float(resolution[0]),
                        float(resolution[1])
                    ]
                },
                "extra": {
                    "source_width": resolution[0],
                    "source_height": resolution[1]
                }
            }
            if caption: data["caption"] = caption
            response = self.session.post(self.api + '/media/configure_to_story/', data={"signed_body": "SIGNATURE." + json.dumps(data)}).json()
            if response['status'] == 'ok': return True
            else: return False
        except:
            return False